@@ -1,5 +1,6 @@
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <inttypes.h>
 #include <sys/stat.h>
 
@@ -118,7 +119,79 @@ double sum(double *v, int n)
   return total;
 }
 
-int forward(const struct TM *ptm, FILE *fin_l, FILE *fout_f, FILE *fout_s)
+int forward_innerloop(size_t pos, const struct TM *ptm,
+    const double *l_curr, const double *f_prev, double *f_curr, double *s_curr)
+{
+  double scaling_factor;
+  double p;
+  double tprob;
+  int isource, isink;
+  int nstates = ptm->nstates;
+  memcpy(f_curr, l_curr, nstates*sizeof(double));
+  if (pos>0)
+  {
+    for (isink=0; isink<nstates; isink++)
+    {
+      p = 0.0;
+      for (isource=0; isource<nstates; isource++)
+      {
+        tprob = ptm->trans[isource*nstates + isink];
+        p += f_prev[isource] * tprob;
+      }
+      f_curr[isink] *= p;
+    }
+  } else {
+    for (isink=0; isink<nstates; isink++)
+    {
+      f_curr[isink] *= ptm->distn[isink];
+    }
+  }
+  scaling_factor = sum(f_curr, nstates);
+  if (scaling_factor == 0.0)
+  {
+    fprintf(stderr, "scaling factor 0.0 at pos %zd\n", pos);
+    return -1;
+  }
+  for (isink=0; isink<nstates; isink++)
+  {
+    f_curr[isink] /= scaling_factor;
+  }
+  *s_curr = scaling_factor;
+  return 0;
+}
+
+int backward_innerloop(size_t pos, const struct TM *ptm,
+    const double *l_prev, const double *l_curr,
+    double scaling_factor, const double *b_prev, double *b_curr)
+{
+  int nstates = ptm->nstates;
+  int isource, isink;
+  int i;
+  double p;
+  if (pos)
+  {
+    for (i=0; i<nstates; i++) b_curr[i] = 0.0;
+    for (isource=0; isource<nstates; isource++)
+    {
+      for (isink=0; isink<nstates; isink++)
+      {
+        p = ptm->trans[isource*nstates + isink];
+        p *= l_prev[isink] * b_prev[isink];
+        b_curr[isource] += p;
+      }
+    }
+  } else {
+    for (i=0; i<nstates; i++) b_curr[i] = 1.0;
+  }
+  for (isource=0; isource<nstates; isource++)
+  {
+    b_curr[isource] /= scaling_factor;
+  }
+  return 0;
+}
+
+int forward_alldisk(const struct TM *ptm,
+    FILE *fin_l, FILE *fout_f, FILE *fout_s)
 {
   /*
    * Run the forward algorithm.
@@ -127,64 +200,39 @@ int forward(const struct TM *ptm, FILE *fin_l, FILE *fout_f, FILE *fout_s)
    * @param fout_f: file of forward vectors
    * @param fout_s: file of scaling factors
    */
+  int errcode = 0;
   int nstates = ptm->nstates;
   double *f_tmp;
+  double *l_curr = malloc(nstates*sizeof(double));
   double *f_curr = malloc(nstates*sizeof(double));
   double *f_prev = malloc(nstates*sizeof(double));
-  if (!f_curr || !f_prev)
+  if (!f_curr || !f_prev || !l_curr)
   {
     fprintf(stderr, "failed to allocate an array\n");
-    return -1;
+    errcode = -1; goto end;
   }
-  double p;
-  double tprob;
-  double scaling_factor;
-  int isource, isink;
+  double sf;
   size_t pos = 0;
-  while (fread(f_curr, sizeof(double), nstates, fin_l))
+  while (fread(l_curr, sizeof(double), nstates, fin_l))
   {
-    /* create the unscaled forward vector */
-    if (pos>0)
+    if (forward_innerloop(pos, ptm, l_curr, f_prev, f_curr, &sf) < 0)
     {
-      for (isink=0; isink<nstates; isink++)
-      {
-        p = 0.0;
-        for (isource=0; isource<nstates; isource++)
-        {
-          tprob = ptm->trans[isource*nstates + isink];
-          p += f_prev[isource] * tprob;
-        }
-        f_curr[isink] *= p;
-      }
-    } else {
-      for (isink=0; isink<nstates; isink++)
-      {
-        f_curr[isink] *= ptm->distn[isink];
-      }
-    }
-    /* scale the forward vector */
-    scaling_factor = sum(f_curr, nstates);
-    if (scaling_factor == 0.0)
-    {
-      fprintf(stderr, "scaling factor 0.0 at pos %lld\n", pos);
-      return -1;
-    }
-    for (isink=0; isink<nstates; isink++)
-    {
-      f_curr[isink] /= scaling_factor;
+      errcode = -1; goto end;
     }
     fwrite(f_curr, sizeof(double), nstates, fout_f);
-    fwrite(&scaling_factor, sizeof(double), 1, fout_s);
+    fwrite(&sf, sizeof(double), 1, fout_s);
     pos++;
-    /* swap f_prev and f_curr */
     f_tmp = f_curr; f_curr = f_prev; f_prev = f_tmp;
   }
+end:
+  free(l_curr);
   free(f_curr);
   free(f_prev);
-  return 0;
+  return errcode;
 }
 
-int backward(const struct TM *ptm, FILE *fin_l, FILE *fin_s, FILE *fout_b)
+int backward_alldisk(const struct TM *ptm,
+    FILE *fin_l, FILE *fin_s, FILE *fout_b)
 {
   /*
    * @param ptm: pointer to the transition matrix struct
@@ -209,35 +257,15 @@ int backward(const struct TM *ptm, FILE *fin_l, FILE *fin_s, FILE *fout_b)
   double *b_prev = malloc(nstates*sizeof(double));
   double *l_curr = malloc(nstates*sizeof(double));
   double *l_prev = malloc(nstates*sizeof(double));
-  int isource, isink;
-  int i;
-  double scaling_factor;
-  double p;
+  double sf;
   size_t pos = 0;
   do
   {
     /* read the likelihood vector and the scaling factor */
     nbytes = fread(l_curr, sizeof(double), nstates, fin_l);
-    nbytes = fread(&scaling_factor, sizeof(double), 1, fin_s);
-    if (pos)
-    {
-      for (i=0; i<nstates; i++) b_curr[i] = 0.0;
-      for (isource=0; isource<nstates; isource++)
-      {
-        for (isink=0; isink<nstates; isink++)
-        {
-          p = ptm->trans[isource*nstates + isink];
-          p *= l_prev[isink] * b_prev[isink];
-          b_curr[isource] += p;
-        }
-      }
-    } else {
-      for (i=0; i<nstates; i++) b_curr[i] = 1.0;
-    }
-    for (isource=0; isource<nstates; isource++)
-    {
-      b_curr[isource] /= scaling_factor;
-    }
+    nbytes = fread(&sf, sizeof(double), 1, fin_s);
+    /* do the inner loop */
+    backward_innerloop(pos, ptm, l_prev, l_curr, sf, b_prev, b_curr);
     /* write the vector */
     fwrite(b_curr, sizeof(double), nstates, fout_b);
     /* swap buffers and increment the position */
@@ -257,7 +285,8 @@ int backward(const struct TM *ptm, FILE *fin_l, FILE *fin_s, FILE *fout_b)
   return 0;
 }
 
-int posterior(int nstates, FILE *fi_f, FILE *fi_s, FILE *fi_b, FILE *fo_d)
+int posterior_alldisk(int nstates,
+    FILE *fi_f, FILE *fi_s, FILE *fi_b, FILE *fo_d)
 {
   /*
    * @param nstates: the number of hidden states
@@ -310,54 +339,27 @@ int forward_somedisk(const struct TM *ptm, FILE *fin_l,
    * @param f_big: the output array of forward vectors
    * @param s_big: the output array of scaling factors
    */
+  int errcode = 0;
   int nstates = ptm->nstates;
-  double p;
-  double tprob;
-  double scaling_factor;
-  int isource, isink;
+  double *l_curr = malloc(nstates*sizeof(double));
   double *s_curr = s_big;
   double *f_curr = f_big;
   double *f_prev = NULL;
   size_t pos=0;
-  while (fread(f_curr, sizeof(double), nstates, fin_l))
+  while (fread(l_curr, sizeof(double), nstates, fin_l))
   {
-    /* create the unscaled forward vector */
-    if (pos>0)
+    if (forward_innerloop(pos, ptm, l_curr, f_prev, f_curr, s_curr) < 0)
     {
-      for (isink=0; isink<nstates; isink++)
-      {
-        p = 0.0;
-        for (isource=0; isource<nstates; isource++)
-        {
-          tprob = ptm->trans[isource*nstates + isink];
-          p += f_prev[isource] * tprob;
-        }
-        f_curr[isink] *= p;
-      }
-    } else {
-      for (isink=0; isink<nstates; isink++)
-      {
-        f_curr[isink] *= ptm->distn[isink];
-      }
+      errcode = -1; goto end;
     }
-    /* scale the forward vector */
-    scaling_factor = sum(f_curr, nstates);
-    if (scaling_factor == 0.0)
-    {
-      fprintf(stderr, "scaling factor 0.0 at pos %zd\n", pos);
-      return -1;
-    }
-    for (isink=0; isink<nstates; isink++)
-    {
-      f_curr[isink] /= scaling_factor;
-    }
-    *s_curr = scaling_factor;
     pos++;
     f_prev = f_curr;
     f_curr += nstates;
     s_curr++;
   }
-  return 0;
+end:
+  free(l_curr);
+  return errcode;
 }
 
 int backward_somedisk(const struct TM *ptm, size_t nobs, FILE *fin_l,
@@ -385,42 +387,17 @@ int backward_somedisk(const struct TM *ptm, size_t nobs, FILE *fin_l,
     errcode = -1; goto end;
   }
   /* start at the end of the scaling factor array */
-  s_curr = s_big + (nobs-1);
+  s_curr = s_big + (nobs - 1);
   /* start at the beginning of the backward array */
   b_curr = b_big;
   size_t nbytes;
-  int isource, isink;
-  int i;
-  double scaling_factor;
   double *ptmp;
-  double p;
   int result;
   size_t pos = 0;
   do
   {
-    /* read the likelihood vector and the scaling factor */
     nbytes = fread(l_curr, sizeof(double), nstates, fin_l);
-    scaling_factor = *s_curr;
-    if (pos)
-    {
-      for (i=0; i<nstates; i++) b_curr[i] = 0.0;
-      for (isource=0; isource<nstates; isource++)
-      {
-        for (isink=0; isink<nstates; isink++)
-        {
-          p = ptm->trans[isource*nstates + isink];
-          p *= l_prev[isink] * b_prev[isink];
-          b_curr[isource] += p;
-        }
-      }
-    } else {
-      for (i=0; i<nstates; i++) b_curr[i] = 1.0;
-    }
-    for (isource=0; isource<nstates; isource++)
-    {
-      b_curr[isource] /= scaling_factor;
-    }
-    /* swap buffers and increment the position */
+    backward_innerloop(pos, ptm, l_prev, l_curr, *s_curr, b_prev, b_curr);
     ptmp = l_curr; l_curr = l_prev; l_prev = ptmp;
     b_prev = b_curr;
     b_curr += nstates;
@@ -443,7 +420,7 @@ int posterior_somedisk(int nstates, size_t nobs,
    * @param nstates: the number of hidden states
    * @param nobs: the number of observations
    * @param f_big: array of forward vectors
-   * @param s_big: array of scaling vectors
+   * @param s_big: array of scaling factors
    * @param b_big: file of backward vectors
    * @param fout_d: file of posterior probability vectors open for writing
    */
@@ -452,7 +429,7 @@ int posterior_somedisk(int nstates, size_t nobs,
   int i;
   const double *f_curr = f_big;
   const double *s_curr = s_big;
-  const double *b_curr = b_big + nstates*(nobs-1);
+  const double *b_curr = b_big + nstates * (nobs - 1);
   /* multiply stuff together and write to the output file */
   for (pos=0; pos<nobs; pos++)
   {
@@ -500,6 +477,137 @@ end:
   return errcode;
 }
 
+int forward_nodisk(const struct TM *ptm, size_t nobs,
+    const double *l_big,
+    double *f_big, double *s_big)
+{
+  /*
+   * Run the forward algorithm.
+   * @param ptm: address of a transition matrix
+   * @param nobs: the number of observations
+   * @param l_big: the input array of likelihood vectors
+   * @param f_big: the output array of forward vectors
+   * @param s_big: the output array of scaling factors
+   */
+  int errcode = 0;
+  int nstates = ptm->nstates;
+  double *s_curr = s_big;
+  double *f_curr = f_big;
+  double *f_prev = NULL;
+  const double *l_curr = l_big;
+  size_t pos;
+  for (pos=0; pos<nobs; pos++)
+  {
+    if (forward_innerloop(pos, ptm, l_curr, f_prev, f_curr, s_curr) < 0)
+    {
+      errcode = -1; goto end;
+    }
+    f_prev = f_curr;
+    f_curr += nstates;
+    s_curr++;
+    l_curr += nstates;
+  }
+end:
+  return errcode;
+}
+
+int backward_nodisk(const struct TM *ptm, size_t nobs,
+    const double *l_big,
+    const double *s_big, double *b_big)
+{
+  /*
+   * @param ptm: pointer to the transition matrix struct
+   * @param nobs: the number of observations
+   * @param l_big: likelihood vectors
+   * @param s_big: scaling factors
+   * @param b_big: backward vectors
+   * @return: negative on error
+   */
+  int nstates = ptm->nstates;
+  const double *s_curr = s_big + (nobs - 1);
+  const double *l_curr = l_big + nstates * (nobs - 1);
+  const double *l_prev = NULL;
+  double *b_curr = b_big;
+  double *b_prev = NULL;
+  size_t pos;
+  for (pos=0; pos<nobs; pos++)
+  {
+    backward_innerloop(pos, ptm, l_prev, l_curr, *s_curr, b_prev, b_curr);
+    l_prev = l_curr;
+    l_curr -= nstates;
+    b_prev = b_curr;
+    b_curr += nstates;
+    s_curr--;
+  }
+  return 0;
+}
+
+int posterior_nodisk(int nstates, size_t nobs,
+    const double *f_big, const double *s_big, const double *b_big,
+    double *d_big)
+{
+  /*
+   * @param nstates: the number of hidden states
+   * @param nobs: the number of observations
+   * @param f_big: array of forward vectors
+   * @param s_big: array of scaling factors
+   * @param b_big: array of backward vectors
+   * @param d_big: array of posterior vectors
+   */
+  size_t pos;
+  int i;
+  const double *f_curr = f_big;
+  const double *s_curr = s_big;
+  const double *b_curr = b_big + nstates * (nobs - 1);
+  double *d_curr = d_big;
+  /* multiply stuff together and write to the output file */
+  for (pos=0; pos<nobs; pos++)
+  {
+    for (i=0; i<nstates; i++)
+    {
+      d_curr[i] = f_curr[i] * s_curr[0] * b_curr[i];
+    }
+    d_curr += nstates;
+    f_curr += nstates;
+    s_curr++;
+    b_curr -= nstates;
+  }
+  return 0;
+}
+
+int fwdbwd_nodisk(const struct TM *ptm, size_t nobs,
+    const double *l_big, double *d_big)
+{
+  int errcode = 0;
+  int nstates = ptm->nstates;
+  /* allocate the big forward, scaling, and backward arrays */
+  double *f_big = malloc(nobs*nstates*sizeof(double));
+  double *s_big = malloc(nobs*sizeof(double));
+  double *b_big = malloc(nobs*nstates*sizeof(double));
+  /* make sure that the arrays were allocated */
+  if (!f_big || !s_big || !b_big)
+  {
+    fprintf(stderr, "failed to allocate an array\n");
+    errcode = -1; goto end;
+  }
+  /* run the forward and backward algorithms */
+  if (forward_nodisk(ptm, nobs, l_big, f_big, s_big) < 0) {
+    errcode = -1; goto end;
+  }
+  if (backward_nodisk(ptm, nobs, l_big, s_big, b_big) < 0) {
+    errcode = -1; goto end;
+  }
+  if (posterior_nodisk(nstates, nobs, f_big, s_big, b_big, d_big) < 0) {
+    errcode = -1; goto end;
+  }
+end:
+  free(f_big);
+  free(s_big);
+  free(b_big);
+  return errcode;
+}
+
+
 int do_fwdbwd_somedisk(const struct TM *ptm,
     const char *likelihoods_name, const char *posterior_name)
 {
@@ -561,7 +669,7 @@ int do_forward(const struct TM *ptm,
     fprintf(stderr, "failed to open the scaling factor file for writing\n");
     errcode = -1; goto end;
   }
-  forward(ptm, fin_l, fout_f, fout_s);
+  forward_alldisk(ptm, fin_l, fout_f, fout_s);
 end:
   fsafeclose(fin_l);
   fsafeclose(fout_f);
@@ -592,7 +700,7 @@ int do_backward(const struct TM *ptm,
     fprintf(stderr, "failed to open the backward vector file for writing\n");
     errcode = -1; goto end;
   }
-  backward(ptm, fin_l, fin_s, fout_b);
+  backward_alldisk(ptm, fin_l, fin_s, fout_b);
 end:
   fsafeclose(fin_l);
   fsafeclose(fin_s);
@@ -629,7 +737,7 @@ int do_posterior(int nstates,
     fprintf(stderr, "failed to open the posterior vector file for writing\n");
     errcode = -1; goto end;
   }
-  posterior(nstates, fin_f, fin_s, fin_b, fout_d);
+  posterior_alldisk(nstates, fin_f, fin_s, fin_b, fout_d);
 end:
   fsafeclose(fin_f);
   fsafeclose(fin_s);
